UNIT VESA;
{$r-,q-,i-,x+,opt+,d-,l-}
{
This is the Vesa part of the graphics subsystem.
------------------------------------------------
Current Version:
         0.11 - 25-02-99 by JacmeT (jacmet@iname.com)

Purpose: 
         To handle the communication between
         the program and the graphics card
         through the VESA 2.0+ interface.
         And if there's support for it uses
         the features of the card (linear
         framebuffer / double/tripple buffering)

Functionality:
         - 32bit protected mode interface support

         (minus indicates not done yet, 
         plus indicates it is implemented and tested
         and question mark indicates implemented
         (perhaps partly) but not tested yet)
           
Comments:
        The use of defines are heavilly approved.
        As of this moment the following defines are
        defined:
        ALL           : All code is defined (Doesn't affect verbose level)
        PMODEINTERFACE: Use Protected mode interface.
        _15BIT        : Support for 15bit modes included.
        _16BIT        : Support for 16bit modes included.
        _24BIT        : Support for 24bit modes included.
        _32BIT        : Support for 32bit modes included.
        VERBOSE_LEVEL0: Verbose level 0 - no messages are included
        VERBOSE_LEVEL1: Verbose level 1 - only the most important messages are included
        VERBOSE_LEVEL2: Verbose level 2 - most messages are included
        VERBOSE_LEVEL3: Verbose level 3 - all messages are included
        DEBUG         : Debugging code included

  
History:
         18-01-99: First revision (0.01)
         --------------------------------
         This is the very first version.
         Not much is working at the moment
         - But it'll come!   

         21-01-99: Second revision (0.02)
         --------------------------------
         Tidied up the code a bit and 
         added more comments / reformatted
         the old comments. 

         24-01-99: Third revision (0.03)
         --------------------------------
         Now we got Init, SetMode and 
         DeInit working - Now the alpha
         test can start - Started work on
         the various flip routines.

         30-01-99: Forth revision (0.04)
         --------------------------------
         24 bit and 32 bit modes now
         working (double buffer) - The next
         thing that I'll implement is the
         Set/Get Display Start function
         (VBE function 07h) to implement
         smooth tripple buffering.
         This will first be done through the
         int 10h interface, and then later
         through the direct pmode interface
         (see PMODEINTERFACE)

         02-02-99: Fifth revision (0.05)
         --------------------------------
         Unfortunately I haven't had time
         to implement the tripple buffering
         code yet, but instead the setmode
         function is changed and a 
         GetSupportedModes procedure is
         addded. This is done because it
         enables us to build a list of
         Supported screenmodes, and therefor
         we can try out larger screenmodes
         than the requested one if the 
         requested screenmode isn't supported.
         (see also the GFX unit.)
         The parameters of the Setmode function
         is changed according to this - now
         it takes a mode number instead of the
         old X,Y and BitsPerPixel. You can get
         the supported mode numbers and their
         resolution and color depth from the
         ModeList variable.
         MemoryModel constants added. This
         is used to see if a given screenmode
         is textmode,cga,packed or a direct
         color mode. (used in GetSupportedModes)

         14-02-99: Sixth revision (0.06)
         --------------------------------
         Minor cosmetic changes. Now all defines
         are done in VESA.INC - int 10h tripple
         buffering is done directly in the FLIP
         unit instead of defining a specific
         function call for it.        

         18-02-99: Seventh revision (0.07)
         ----------------------------------
         the X and Y variables (virtual X and Y-
         resolution) moved to FLIP.PAS to 
         accommodate the need of the new scaling
         version of the VGA flip procedure.

         22-02-99: Eighth revision (0.08)
         ---------------------------------
         Included HARDWARE.INC - this is
         done to clean up the include
         files, and collect all the hardware
         related defines in 1 file, and
         it also lets VESA.PAS know which
         color depths are defined, and hence
         the GetSupportedModes procedure is
         changes, so that only defined 
         color depths are included in the list.
         This fixes the bug introduced in 
         version 0.07 of GFX.PAS.

         23-02-99: Ninth revision (0.09)
         --------------------------------
         Added debug support. This should speed up
         remote debugging / beta testing. This way
         all data of the system is written to a file
         which can then be examined later. This way
         users who discovers a bug can send me a copy
         of the file, and it should help debugging
         quite a lot.

         24-02-99: Tenth revision (0.10)
         --------------------------------
         Turned on range checking and found
         a bug in GetSupportedModes. It was
         supposed to check every mode from
         0x100 to 0x1FF, but actually it 
         checked mode 0x200 as well and thus
         overwriting parts of memory that it
         shouldn't touch. Started looking at
         what causes the system to generate
         protection faults under pure dos
         (which it didn't do in the early
         versions) - Possible found a bug in
         part of the verbose output of 
         VESA.Init, but I'll have to look
         into that furter.

         25-02-99: Eleventh revision (0.11)
         -----------------------------------
         Bugfix release.
         Found a bug in the Setmode procedure
         related to the mapping of the 
         framebuffer to logical memory (DPMI 
         function 0x0800) - This is what
         caused the general protection faults
         under pure dos, because the function
         expects the nuber of bytes to map,
         and I sent it the number of Kbytes.
         This way only the first page (4kb)
         was mapped, and an GPF would occur 
         as soon as the rest of the
         framebuffer was accessed. Strangely
         enough windows didn't seem to have
         any problems with it though.
}

{$I VESA.INC}
{$I HARDWARE.INC}
{$I ..\GENERAL.INC}

{$IFDEF ALL }
 {$DEFINE _VESA }
 {$DEFINE _VGA }
 {$DEFINE _RGB }
 {$DEFINE _BGR }
 {$DEFINE _32BIT }
 {$DEFINE _24BIT }
 {$DEFINE _16BIT }
 {$DEFINE _15BIT }
 {$DEFINE TRIPPLEBUFFER }
 {$DEFINE SCALING }
 {{$DEFINE PMODEINTERFACE }
 {$DEFINE WAITFORRETRACE }
{$ENDIF }

{$IFDEF _32BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _24BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _16BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _15BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFNDEF _BGR }
 {$DEFINE _RGB }
{$ENDIF }

{$IFNDEF _RGB }
 {$DEFINE _BGR }
{$ENDIF }

INTERFACE

{$IFDEF DEBUG}
 USES DPMI,DEBUGGER,DEBUG;
{$ELSE}
 USES DPMI; {to simulate realmode interrupt}
{$ENDIF}

TYPE
    ControllerInfoType = Record
                Signature    : array[1..4] of char; {VESA}
                MinVersion   : byte;
                MajVersion   : byte;
                Oemptr       : pchar;
                Capabil      : dword;
                Modeptr      : ^array[0..1] of word;
                Memory       : word; {number of 64kb blocks}
                MinSoftRev   : byte;
                MajSoftRev   : byte;
                Vendnameptr  : pchar;
                Prodnameptr  : pchar;
                Prodrevptr   : pchar;
                Reserved     : array[1..222] of byte;
                Oemdata      : array[1..256] of byte;
                         end;

    ModeInfoType       = Record
                {Information for all VBE versions}
                Modeattr     : word;
                WinAattr     : byte;
                WinBattr     : byte;
                Wingranular  : word;
                Winsize      : word;
                WinAsegment  : word;
                WinBsegment  : word;
                Winfuncptr   : pointer;
                Bytesperscan : word;
                {Information for VBE 1.2 and above}
                Xres         : word;
                Yres         : word;
                Xcharsize    : byte;
                Ycharsize    : byte;
                Numofplanes  : byte;
                Bitsperpixel : byte;
                Numofbanks   : byte;
                Memorymodel  : byte;
                Banksize     : byte;
                NumofImages  : byte;
                Reserved     : byte;
                {Direct color fields (Required for direct/6 and YUV/7 memory models)}
                Redmasksize  : byte;
                Redfieldpos  : byte;
                Greenmasksize: byte;
                Greenfieldpos: byte;
                Bluemasksize : byte;
                Bluefieldpos : byte;
                Resvdmasksize: byte;
                Resvdfieldpos: byte;
                Dircolmodinfo: byte;
                {Information for VBE 2.0 and above}
                Physbaseptr  : pointer;
                Offscrnmemofs: dword;
                Offscrnmemsiz: word; {number of kb}
                Reserved2    : array[1..206] of byte;
                         end;

    ModeType           = Record
                X,Y          : word;
                BitsPerPixel : word;
                ModeNumber   : word;           
                         end; 

{$IFDEF PMODEINTERFACE}
VAR PmodeInterFace: record
                     DisplayWindowControl: ^procedure;
                     SetDisplayStart     : ^procedure;
                     SetPalette          : ^procedure;
                    end; 
{$ENDIF}

VAR ControllerInfo: VESA.ControllerInfoType;
    ModeInfo      : VESA.ModeInfoType;
    ModeList      : array[0..255] of VESA.ModeType;

Function Init: Boolean;
{ The Initialization function.}
Procedure GetSupportedModes;
{ The GetSupportedModes procedure.}
Function Setmode(ModeNumber: dword): boolean;
{ The Setmode function. }
Procedure DeInit;
{ The DeInitialization procedure.}

IMPLEMENTATION

Const {constants used for MemoryModel - see ModeInfoType}
     _MemoryModel_TextMode  = $00;
     _MemoryModel_CGA       = $01;
     _MemoryModel_Hercules  = $02;
     _MemoryModel_Planar    = $03;
     _MemoryModel_Packed    = $04;
     _MemoryModel_NonChain4 = $05;
     _MemoryModel_DirectCol = $06;
     _MemoryModel_YUV       = $07;

Function Init: Boolean;
{ The initialization function. 
  -----------------------------
Purpose:
         This function detects if
         there is a VESA 2.0+ 
         compatible interface 
         available, and sets up
         the VESA subsystem.

Comments:
         This function must be called
         before any of the others -
         See also the DeInit procedure
         The function returns a boolean
         expression telling if the 
         function call was successful
         or not.
}         
 var Temp_ControllerInfo: ^VESA.ControllerInfoType; {temporary pointer for VESA controller Info}
 {$IFDEF PMODEINTERFACE}    
 var address: dword; {temporary address of 32bit interface start}
 {$ENDIF}

 begin
  {Return false if error}
  Init:=False;
  
  {allocate realmode memory through DPMI for VESA call and point Temp_ControllerInfo^ to it}
  Temp_ControllerInfo:=DPMI.AllocateRealmodeMemory(sizeof(VESA.ControllerInfoType) shr 4);
  {if Temp_ControllerInfo=nil then memory allocation failed}
  if Temp_ControllerInfo=nil then exit;  

  Temp_ControllerInfo^.Signature:='VBE2';   {to get VBE 2.0 info.      }
  
  DPMI.RealmodeRegisters._edi:=0;
  DPMI.RealmodeRegisters._es:=word(dword(Temp_ControllerInfo) shr 4);
  DPMI.RealmodeRegisters._eax:=$4F00;
  DPMI.RealmodeRegisters._sp:=0;
  DPMI.RealmodeRegisters._ss:=0;
  {eax=4F00h => VESA 'Return VBE Controller Information' function call}

  {was function call succesfull and VESA version >= 2.0 ?}
  if (not(DPMI.SimulateRealmodeInterrupt($10)) or 
      not((DPMI.RealmodeRegisters._eax and $FFFF) = $4F) or
      not(Temp_ControllerInfo^.Signature='VESA') or
      not(Temp_ControllerInfo^.MajVersion>1)) then 
   begin
    DPMI.DeAllocateRealmodeMemory;
    {$IFDEF VERBOSE_LEVEL2}
     If not(((DPMI.RealmodeRegisters._eax and $FFFF) = $4F) or
        (Temp_ControllerInfo^.Signature='VESA')) then 
      begin 
       Writeln('VESA: Interface Not detected!');
      end
     else 
      begin
       Writeln('VESA: Version 2.0 or greater is required! (',
                Temp_ControllerInfo^.MajVersion,'.',
                Temp_ControllerInfo^.MinVersion,' detected)');
       {$IFDEF DEBUG}
        Writeln(DEBUGGER.FileHandle,'VESA: Version ',Temp_ControllerInfo^.MajVersion,'.',Temp_ControllerInfo^.MinVersion,' detected');
       {$ENDIF}       
      end;
    {$ENDIF}
    exit;
   end;

  {copy data fra realmode memory to protected mode mem.}
  ControllerInfo:=Temp_ControllerInfo^; 

  {$IFDEF VERBOSE_LEVEL2}
   {convert pointers from realmode to protected mode}
   dword(ControllerInfo.OemPtr)     :=dword(@ControllerInfo) + (dword(ControllerInfo.OemPtr)      and $FFFF);
   dword(ControllerInfo.VendnamePtr):=dword(@ControllerInfo) + (dword(ControllerInfo.VendnamePtr) and $FFFF); 
   dword(ControllerInfo.ProdnamePtr):=dword(@ControllerInfo) + (dword(ControllerInfo.ProdnamePtr) and $FFFF);
   dword(ControllerInfo.ProdrevPtr) :=dword(@ControllerInfo) + (dword(ControllerInfo.ProdrevPtr)  and $FFFF);
  {$ENDIF}

  {$IFDEF VERBOSE_LEVEL1}
   Writeln('VESA: Version ',ControllerInfo.MajVersion,'.',ControllerInfo.MinVersion,' interface with ',ControllerInfo.Memory shl 6,' Kb detected');
   {$IFDEF VERBOSE_LEVEL2}
    Writeln('VESA: Software revision: ',ControllerInfo.MajSoftRev,'.',ControllerInfo.MinSoftRev);
    Writeln('VESA: OEM name   : ',ControllerInfo.Oemptr);
    Writeln('VESA: Vendor name: ',ControllerInfo.Vendnameptr);
    Writeln('VESA: Prod. name : ',ControllerInfo.Prodnameptr);
    Writeln('VESA: Prod. rev. : ',ControllerInfo.Prodrevptr);
    {$IFDEF DEBUG}
     Writeln(DEBUGGER.FileHandle,'VESA: Version ',ControllerInfo.MajVersion,'.',ControllerInfo.MinVersion,' interface with ',ControllerInfo.Memory shl 6,' Kb detected');
     Writeln(DEBUGGER.FileHandle,'VESA: Software revision: ',ControllerInfo.MajSoftRev,'.',ControllerInfo.MinSoftRev);
     Writeln(DEBUGGER.FileHandle,'VESA: OEM name   : ',ControllerInfo.Oemptr);
     Writeln(DEBUGGER.FileHandle,'VESA: Vendor name: ',ControllerInfo.Vendnameptr);
     Writeln(DEBUGGER.FileHandle,'VESA: Prod. name : ',ControllerInfo.Prodnameptr);
     Writeln(DEBUGGER.FileHandle,'VESA: Prod. rev. : ',ControllerInfo.Prodrevptr);
    {$ENDIF}
   {$ENDIF}
  {$ENDIF}

  {free realmode memory block}
  DPMI.DeAllocateRealmodeMemory;  

  {$IFDEF PMODEINTERFACE}
   {get protected mode interface (function 4F0Ah)}
   DPMI.RealmodeRegisters._eax:=$4F0A;
   DPMI.RealmodeRegisters._ebx:=0;
   DPMI.RealmodeRegisters._sp:=0;
   DPMI.RealmodeRegisters._ss:=0;
  {$ENDIF}
  Init:=True;
 end; {Init}

Procedure GetSupportedModes;
{ The GetSupportedModes procedure. 
  --------------------------------
Purpose:
         This procedure detects which
         (direct color) screenmodes
         are supported, and fills
         the ModeList array.

Comments:
         The procedure runs through every
         possible Vesa mode and checks if
         the given mode is available. This
         is done (instead of just checking
         the supported modes in 
         ControllerInfoType.ModePtr^)
         because of buggy VBE 
         implementations on some cards.

         The mode numbers in the ModeList
         Should be used as parameters
         for the Setmode function.
         The ModeList will be sorted this
         way: 
         Going from smallest to biggest
         X- and  Y-resolution, and going
         from Biggest screen depth to 
         smallest screen depth.

         Example:
         The modes 320x200x24,320x200x32,
         320x240x32,640x480x16,640x480x24

         Will be sorted like this:
         320x200x32,320x200x24,320x240x32,
         640x480x24,640x480x16.

         22-02-99: Now only the modes defined
         in HARDWARE.INC is put in the list.
}         
 var Temp_ModeList : array[0..255] of VESA.ModeType; {temporary list of (unsorted) supported modes}
     Temp_ModeInfo :^VESA.ModeInfoType; {temporary pointer for VESA Mode Info}
     CurrentMode   : dword;
     X,Y,BitsPerPixel,ModeNumber,Count,SortCount: dword; {needed for sorting}
 begin

  {allocate realmode memory through DPMI for VBE call and point Temp_ModeInfo^ to it}
  Temp_ModeInfo:=DPMI.AllocateRealmodeMemory(sizeof(ModeInfoType) shr 4);
  {if Temp_ModeInfo=nil then memory allocation failed}
  if Temp_ModeInfo=nil then exit;

  DPMI.RealmodeRegisters._edi:=0;
  DPMI.RealmodeRegisters._es:=word(dword(Temp_ModeInfo) shr 4); {segment of Temp_ModeInfo}
  DPMI.RealmodeRegisters._sp:=0;
  DPMI.RealmodeRegisters._ss:=0; 

  {first VESA mode defined is 100h (last is 1FFh)}
  Currentmode:=$FF;
  
  Repeat 
   inc(Currentmode);
   DPMI.RealmodeRegisters._ecx:=Currentmode;
   DPMI.RealmodeRegisters._eax:=$4F01;
   {eax = 4F01h => VESA 'Return VBE mode information' function call}
   if (DPMI.SimulateRealmodeInterrupt($10) and 
      ((DPMI.RealmodeRegisters._eax and $FFFF) = $4F)
      and (Temp_ModeInfo^.MemoryModel = _MemoryModel_DirectCol)) then
    begin
     {mode supported}
     {$IFDEF VERBOSE_LEVEL2}
      Writeln('VESA: Mode ',CurrentMode,' (',Temp_ModeInfo^.Xres,'x',Temp_ModeInfo^.Yres,'x',Temp_ModeInfo^.BitsPerPixel,') supported');
     {$ENDIF}

     {$IFDEF DEBUG}
      Writeln(DEBUGGER.FileHandle,'VESA: Mode ',CurrentMode,' Supported');
      Writeln(DEBUGGER.FileHandle,'VESA: Resolution: ',Temp_ModeInfo^.Xres,'x',Temp_ModeInfo^.Yres,'x',Temp_ModeInfo^.BitsPerPixel);
     {$ENDIF}     
     {add supported mode to Temp_ModeList}
     Temp_ModeList[CurrentMode-$100].X:=Temp_ModeInfo^.Xres;
     Temp_ModeList[CurrentMode-$100].Y:=Temp_ModeInfo^.Yres;
     Temp_ModeList[CurrentMode-$100].BitsPerPixel:=Temp_ModeInfo^.BitsPerPixel;
     {$IFNDEF _32BIT}
      {if 32 bit modes not defined, then skip those}
      if Temp_ModeList[CurrentMode-$100].BitsPerPixel=32 then
       begin
        Temp_ModeList[CurrentMode-$100].X:=$FFFF;        {because we sort from small to large}
        Temp_ModeList[CurrentMode-$100].Y:=$FFFF;
        Temp_ModeList[CurrentMode-$100].BitsPerPixel:=0; {because we sort from large to small}
       end;
     {$ENDIF}
     {$IFNDEF _24BIT}
      {if 24 bit modes not defined, then skip those}
      if Temp_ModeList[CurrentMode-$100].BitsPerPixel=24 then
       begin
        Temp_ModeList[CurrentMode-$100].X:=$FFFF;        {because we sort from small to large}
        Temp_ModeList[CurrentMode-$100].Y:=$FFFF;
        Temp_ModeList[CurrentMode-$100].BitsPerPixel:=0; {because we sort from large to small}
       end;
     {$ENDIF}
     {$IFNDEF _16BIT}
      {if 16 bit modes not defined, then skip those}
      if Temp_ModeList[CurrentMode-$100].BitsPerPixel=16 then
       begin
        Temp_ModeList[CurrentMode-$100].X:=$FFFF;        {because we sort from small to large}
        Temp_ModeList[CurrentMode-$100].Y:=$FFFF;
        Temp_ModeList[CurrentMode-$100].BitsPerPixel:=0; {because we sort from large to small}
       end;
     {$ENDIF}
     {$IFNDEF _15BIT}
      {if 15 bit modes not defined, then skip those}
      if Temp_ModeList[CurrentMode-$100].BitsPerPixel=15 then
       begin
        Temp_ModeList[CurrentMode-$100].X:=$FFFF;        {because we sort from small to large}
        Temp_ModeList[CurrentMode-$100].Y:=$FFFF;
        Temp_ModeList[CurrentMode-$100].BitsPerPixel:=0; {because we sort from large to small}
       end;
     {$ENDIF}
    end
   else
    begin
     {mode not supported}
     {$IFDEF VERBOSE_LEVEL4}
      Writeln('VESA: Mode ',CurrentMode,' not supported');
     {$ENDIF}
     {fill entry in Temp_ModeList with illegal values}
     Temp_ModeList[CurrentMode-$100].X:=$FFFF; {because we sort from small to large}
     Temp_ModeList[CurrentMode-$100].Y:=$FFFF;
     Temp_ModeList[CurrentMode-$100].BitsPerPixel:=0; {because we sort from large to small}
    end;
   Temp_ModeList[CurrentMode-$100].ModeNumber:=CurrentMode;
  until Currentmode=$1FF; {maximum modenumber is 1FFh}

  {sort in following order. Xsize/Ysize is from small to large. ColorDepth is 32/24/16/15}
  SortCount:=0;

  repeat
   X:=$FFFF; Y:=$FFFF; BitsPerPixel:=0; ModeNumber:=$FFFF; 
   for Count:=0 to 255 do
    begin
     If ((Temp_ModeList[Count].X<X) or
        (Temp_ModeList[Count].Y<Y)) then
      begin
       {current resolution is smaller}
       X:=Temp_ModeList[Count].X;
       Y:=Temp_ModeList[Count].Y;
       BitsPerPixel:=Temp_ModeList[Count].BitsPerPixel;
       ModeNumber:=Temp_ModeList[Count].ModeNumber;
      end;
    
     If ((Temp_ModeList[Count].X=X) and
        (Temp_ModeList[Count].Y=Y) and 
        (Temp_ModeList[Count].BitsPerPixel>BitsPerPixel)) then
      begin
       {current color depth is larger}
       {no need to change X and Y - they are the same}
       BitsPerPixel:=Temp_ModeList[Count].BitsPerPixel;
       ModeNumber:=Temp_ModeList[Count].ModeNumber;
      end; 
    end;
   if (ModeNumber<$FFFF) then
    begin
     {found a new mode}

     {$IFDEF VERBOSE_LEVEL3}
      Writeln('VESA: Mode ',ModeNumber,' (',X,'x',Y,'x',BitsPerPixel,') inserted in ModeList at position ',SortCount);
     {$ENDIF}

     {add mode to sorted list (ModeList)}
     ModeList[SortCount].X:=X;
     ModeList[SortCount].Y:=Y;
     ModeList[SortCount].BitsPerPixel:=BitsPerPixel;
     ModeList[SortCount].ModeNumber:=ModeNumber;

     {remove mode from temporary list}
     Temp_ModeList[ModeNumber-$100].X:=$FFFF;
     Temp_ModeList[ModeNumber-$100].Y:=$FFFF;
     Temp_ModeList[ModeNumber-$100].BitsPerPixel:=0;

     {next entry in sorted list}
     SortCount:=SortCount+1;
    end;
  until ModeNumber=$FFFF; {repeat until all modes are sorted}

  for SortCount:=SortCount to 255 do
   begin
    {fill rest of ModeList with illegal values 
     - actually it's only necessary to fill 
    the next mode with illegal values, because
    in the GFX unit the search will stop when 
    an illegal mode is encountered. There could
    be a bug here if the VESA interface 
    supported 256 direct color modes, but that
    is not very likely - because then there 
    wouldn't be any room for any other modes
    like textmode of 256 color modes}
    ModeList[SortCount].X:=$FFFF;
    ModeList[SortCount].Y:=$FFFF;
    ModeList[SortCount].BitsPerPixel:=0;
    ModeList[SortCount].ModeNumber:=$FFFF;
   end; 
  {free realmode memory block}
  DPMI.DeAllocateRealmodeMemory;
 end; {GetSupportedModes}

function Setmode(ModeNumber: dword): boolean;
{ The SetMode function. 
  -----------------------------
Purpose:
         This function tries to
         set the given screenmode
         and map a pointer to it.
         Further more it fills the
         ModeInfo variable with
         info on the screen mode.

Comments:
         See also the modeinfo variable
         and the ModeList array.
         This function only supports
         linear framebuffer, not banked
         modes, and will therefor fail if
         such isn't available.
         The function returns a boolean
         expression telling if the 
         function call was successful
         or not.
}         
 var Temp_ModeInfo:^VESA.ModeInfoType; {temporary pointer for VESA Mode Info}
     result: dword;
 begin
  Setmode:=False;
  asm
   mov eax, 04F02h           {VESA 'Set VBE Mode' function call}
   mov ebx,[ModeNumber]
   add ebx,0100000000000000b {linear frame buffer / clear display memory}
   int 10h
   mov [result],eax          {if function call failed then result != 0x4Fh}
  end;  
  if result<>$04F then 
   begin
    {mode set falied}
    {$IFDEF VERBOSE_LEVEL2}
     Writeln('VESA: Mode ',ModeList[ModeNumber-$100].X,'x',ModeList[ModeNumber-$100].Y,'x',ModeList[ModeNumber-$100].BitsPerPixel,'xLFB mode set failed');
    {$ENDIF}
    exit;
   end;

  {get modeinfo}

  {allocate realmode memory through DPMI for VBE call and point Temp_ModeInfo^ to it}
  Temp_ModeInfo:=DPMI.AllocateRealmodeMemory(sizeof(ModeInfoType) shr 4);
  {if Temp_ModeInfo=nil then memory allocation failed}
  if Temp_ModeInfo=nil then exit;

  DPMI.RealmodeRegisters._edi:=0;
  DPMI.RealmodeRegisters._es:=word(dword(Temp_ModeInfo) shr 4); {segment of Temp_ModeInfo}
  DPMI.RealmodeRegisters._sp:=0;
  DPMI.RealmodeRegisters._ss:=0; 
  DPMI.RealmodeRegisters._eax:=$4F01; {eax = 4F01h => VESA 'Return VBE mode information' function call}
  DPMI.RealmodeRegisters._ecx:=ModeNumber;

  if not(DPMI.SimulateRealmodeInterrupt($10) and 
     ((DPMI.RealmodeRegisters._eax and $FFFF) = $4F)) then exit; {interrupt failed?}

  {copy data fra realmode memory to protected mode mem.}
  ModeInfo:=Temp_ModeInfo^;

  {$IFDEF DEBUG}
   Writeln(DEBUGGER.FileHandle,'VESA: Setmode ',ModeNumber,' called');
   Writeln(DEBUGGER.FileHandle,'VESA: Mode Supported  : ',ModeInfo.ModeAttr and $01);
   Writeln(DEBUGGER.FileHandle,'VESA: Graphics mode   : ',(ModeInfo.ModeAttr and $10) shr 4);
   Writeln(DEBUGGER.FileHandle,'VESA: Vga compatible  : ',(ModeInfo.ModeAttr and $20) shr 5);
   Writeln(DEBUGGER.FileHandle,'VESA: Linear framebuf : ',(ModeInfo.ModeAttr and $80) shr 7);
   Writeln(DEBUGGER.FileHandle,'VESA: Resolution      : ',ModeInfo.Xres,'x',ModeInfo.Yres,'x',ModeInfo.BitsPerPixel);
   Writeln(DEBUGGER.FileHandle,'VESA: BytesPerScanline: ',ModeInfo.BytesPerScan);
   Writeln(DEBUGGER.FileHandle,'VESA: Direct color (6): ',ModeInfo.MemoryModel);
   Writeln(DEBUGGER.FileHandle,'VESA: Number of pages : ',ModeInfo.NumOfImages+1);
   Writeln(DEBUGGER.FileHandle,'VESA: Red   : ',ModeInfo.RedMaskSize,' bits at ',ModeInfo.RedFieldPos);
   Writeln(DEBUGGER.FileHandle,'VESA: Green : ',ModeInfo.GreenMaskSize,' bits at ',ModeInfo.GreenFieldPos);
   Writeln(DEBUGGER.FileHandle,'VESA: Blue  : ',ModeInfo.BlueMaskSize,' bits at ',ModeInfo.BlueFieldPos);
   Writeln(DEBUGGER.FileHandle,'VESA: Programmable DAC: ',ModeInfo.DirColModInfo and $01);
   Writeln(DEBUGGER.FileHandle,'VESA: Rsvd Bits usable: ',(ModeInfo.DirColModInfo and $02) shr 1);
   Writeln(DEBUGGER.FileHandle,'VESA: Physical address: ',dword(ModeInfo.Physbaseptr));
   Writeln(DEBUGGER.FileHandle,'VESA: Offscreen Memory: ',ModeInfo.OffscrnmemSiz,' kb at ',ModeInfo.Offscrnmemofs);
  {$ENDIF} 

  {free realmode memory block}
  DPMI.DeAllocateRealmodeMemory;

  {map physical memory (framebuffer) with DPMI}
  ModeInfo.PhysBasePtr:=DPMI.MapPhysicalMemory(ModeInfo.PhysBasePtr,ControllerInfo.Memory shl 16);
  {if ModeInfo.PhysBasePtr=nil then memory mapping failed}
  If ModeInfo.PhysBasePtr=nil then exit;

  {$IFDEF DEBUG}
   Writeln(DEBUGGER.FileHandle,'VESA: ',ControllerInfo.Memory shl 6,' kb of physical address mapped to: ',dword(ModeInfo.Physbaseptr));
  {$ENDIF}
  Setmode:=True;
 end; {SetMode}

Procedure DeInit;
{ The deinitialization function. 
  ------------------------------
Purpose:
         This procedure deinitializes
         the VESA interface and returns
         to standard mode 3 textmode.
Comments:
         This procedure must be called
         After using the VESA interface
         - See also the Init function
         This procedure tries to unmap
         the framebuffer through DPMI.
}         
 begin
  DPMI.FreeMemoryMapping(ModeInfo.PhysBasePtr);
  asm
   mov ax,3
   int 10h
  end;
  {$IFDEF VERBOSE_LEVEL2}
   Writeln('VESA: DeInitialization successful');
  {$ENDIF}
 end; {DeInit}

BEGIN
END.
